/*
 * thread.c: The thread for the Atmolight-plugin
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */
#include <stdlib.h>
#include <linux/videodev.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <arpa/inet.h>
#include <vdr/tools.h>
#include <vdr/thread.h>
#include "setup.h"
#include "thread.h"

// macros
#define MIN(X, Y)  ((X) < (Y) ? (X) : (Y))
#define MAX(X, Y)  ((X) > (Y) ? (X) : (Y))

// capture width/height
#define WIDTH    64
#define HEIGHT   48

// for sending data
extern sockaddr_in address;
extern char *arg_serial;

// --- cAtmoV4L_Capture ------------------------------------------------------

class cAtmoV4L_Capture {
private:
  struct video_mmap v_map;
  void *map;
  int fd;
public:
  cAtmoV4L_Capture(int width, int height);
  ~cAtmoV4L_Capture();
  void *Capture();
};

cAtmoV4L_Capture::cAtmoV4L_Capture(int width, int height)
{
  int res;
  if ((fd=open("/dev/video0",O_RDWR)) == -1)
  {
    perror(NULL);
    return;
  }

  struct video_capability cap;
  ioctl(fd,VIDIOCGCAP,&cap);

  struct video_window win;
  res = ioctl(fd,VIDIOCGWIN,&win);
  win.x=win.y=win.chromakey=0;
  win.width=width;
  win.height=height;
  win.flags=0;
  res = ioctl(fd,VIDIOCSWIN,&win);

  struct video_picture pic;
  res = ioctl(fd,VIDIOCGPICT,&pic);

  pic.depth=24;
  pic.palette=VIDEO_PALETTE_RGB24;
  pic.brightness=pic.hue=pic.colour=pic.contrast=pic.whiteness=32767;
  res = ioctl(fd,VIDIOCSPICT,&pic);

  struct video_mbuf m_buf;
  res = ioctl(fd,VIDIOCGMBUF,&m_buf);

  map=mmap(0,m_buf.size,PROT_READ | PROT_WRITE,MAP_SHARED,fd,0);

  v_map.frame=0;
  v_map.width=width;
  v_map.height=height;
  v_map.format=VIDEO_PALETTE_RGB24;
}

cAtmoV4L_Capture::~cAtmoV4L_Capture()
{
 close(fd);
}

void *cAtmoV4L_Capture::Capture()
{
 ioctl(fd,VIDIOCMCAPTURE,&v_map);
 ioctl(fd,VIDIOCSYNC,&v_map);
 return map;
}

// --- cAtmoThread -----------------------------------------------------------

cAtmoThread::cAtmoThread()
{
 /* serial communication */
 dev_handle = -1;
 if (arg_serial != NULL)
 { 
   dev_handle = open(arg_serial, O_RDWR | O_NOCTTY);
   if (dev_handle >= 0)
   {
     int bconst = B57600;
     struct termios tio;
     memset(&tio,0,sizeof(tio));
     tio.c_cflag = (CS8 | CREAD | HUPCL | CLOCAL);
     tio.c_iflag = (INPCK | BRKINT);
     tio.c_cc[VMIN] = 1; // read 1 character blocking
     cfsetispeed(&tio, bconst);
     cfsetospeed(&tio, bconst);

     if(!tcsetattr(dev_handle, TCSANOW, &tio))
     {
       CHECK(tcflush(dev_handle, TCIOFLUSH)); // Flush
     }
     else
     {
       close(dev_handle);
       dev_handle = -1;
       esyslog("tcsetattr failed: %s",strerror(errno));
     }
   }
   if (0 > dev_handle) { esyslog("Atmolight: Could not open serial port!"); }
   else { esyslog("Atmolight: Serial port open!"); }
 }
}

cAtmoThread::~cAtmoThread()
{
 char cmd[4]; 
 if (Running()) { Cancel(1); }

 /* --- create string to send (all off) --- */
 sprintf(cmd, "%c%c%c%c", 0x80, 0x00, 0x00, 0x00);
 write(dev_handle, cmd, 4);

 /* --- close serial port --- */
 if (0 <= dev_handle) { close(dev_handle); }
}

void cAtmoThread::Action(void)
{
 cTimeMs t;
 char cmd[4];
 tRGBColor color;

 unsigned char read_buf[1];
 unsigned int setpoint_red, setpoint_green, setpoint_blue; // lut-entry
 unsigned int ad_val_red, ad_val_green, ad_val_blue; // measured AD value

 FILE *fd_w;
 fd_w = fopen("/tmp/adval", "w");

 cCondWait::SleepMs(1000); // wait 1,0 second

 AtmoSetup.GetStartMode(); // manual or automatic start of vdr?

 while (Running())
 {
   t.Set(0);
   /* --- */

     if (true == AtmoSetup.atmo_on)
     {
       switch (AtmoSetup.Mode)
       {
         case 0: // live mode
                 color = CalcColors(); 
                 // create string to send
                 cmd[0] = 0x80; // startbyte
                 cmd[1] = color.r;
                 cmd[2] = color.g;
                 cmd[3] = color.b;
         break;

         case 1: // static mode
                 // create string to send
                 cmd[0] = 0x80; // startbyte
                 cmd[1] = AtmoSetup.StaticRed;
                 cmd[2] = AtmoSetup.StaticGreen;
                 cmd[3] = AtmoSetup.StaticBlue;
         break;

         case 2: // white calibration
                 // create string to send
                 cmd[0] = 0x80; // startbyte
                 cmd[1] = AtmoSetup.CalibRed;
                 cmd[2] = AtmoSetup.CalibGreen;
                 cmd[3] = AtmoSetup.CalibBlue;
         break;

         case 3: // raw mode
         default: 
                 // create string to send 
                 cmd[0] = 0x80 | 0x01 | 0x02; // startbyte, raw PWM mode, request AD-data
                 cmd[0] |= (AtmoSetup.RawRed & 0x80) >> 1;
                 cmd[0] |= (AtmoSetup.RawGreen & 0x80) >> 2;
                 cmd[0] |= (AtmoSetup.RawBlue & 0x80) >> 3;
                 cmd[1] = AtmoSetup.RawRed & 0x7F;
                 cmd[2] = AtmoSetup.RawGreen & 0x7F;
                 cmd[3] = AtmoSetup.RawBlue & 0x7F;
         break;
       }
     }
     else // (false == AtmoSetup.atmo_on)
     {     
       cmd[0] = 0x80; // startbyte
       cmd[1] = 0x00;
       cmd[2] = 0x00;
       cmd[3] = 0x00;
     }

   /* --- send data to target: over serial port --- */
   if (0 <= dev_handle) 
   {
     write(dev_handle, cmd, 4);
     CHECK(tcflush(dev_handle, TCIOFLUSH)); // Flush
   }

   /* --- send data to target: over network --- */
   if (0 != address.sin_port)
   {
     int create_socket;
     if ((create_socket = socket(AF_INET, SOCK_STREAM, 0)) > 0)
     {
       if (0 >= connect(create_socket, (struct sockaddr *) &address, sizeof(address)))
       {
         send(create_socket, cmd, 4, 0);
         close(create_socket);
       }
     }
   }

   /* read back data from atmolight */
   if (AtmoSetup.Mode == 3)
   {
     do {
       read(dev_handle, read_buf, 1);
     } while (read_buf[0] != 0xFF);

     do {
       read(dev_handle, read_buf, 1);
     } while (read_buf[0] != 0xFE);
 
     read(dev_handle, read_buf, 1);   
     setpoint_red = (read_buf[0] << 8);
     read(dev_handle, read_buf, 1);   
     setpoint_red |= (read_buf[0]);

     read(dev_handle, read_buf, 1);   
     ad_val_red = (read_buf[0] << 8);
     read(dev_handle, read_buf, 1);   
     ad_val_red |= (read_buf[0]);

     read(dev_handle, read_buf, 1);   
     setpoint_green = (read_buf[0] << 8);
     read(dev_handle, read_buf, 1);   
     setpoint_green |= (read_buf[0]);

     read(dev_handle, read_buf, 1);   
     ad_val_green = (read_buf[0] << 8);
     read(dev_handle, read_buf, 1);   
     ad_val_green |= (read_buf[0]);

     read(dev_handle, read_buf, 1);   
     setpoint_blue = (read_buf[0] << 8);
     read(dev_handle, read_buf, 1);   
     setpoint_blue |= (read_buf[0]);

     read(dev_handle, read_buf, 1);   
     ad_val_blue = (read_buf[0] << 8);
     read(dev_handle, read_buf, 1);   
     ad_val_blue |= (read_buf[0]);

     fprintf(fd_w, "Red: %d, Green: %d, Blue: %d\n", ad_val_red, ad_val_green, ad_val_blue);
     fflush(fd_w);
   }

   /* --- */
   // esyslog("%d",t.Elapsed()); // print duration
   cCondWait::SleepMs(AtmoSetup.SleepTime - t.Elapsed());
 }

 fclose(fd_w);
}

tRGBColor cAtmoThread::CalcColors(void)
{
 static cAtmoV4L_Capture cap(WIDTH, HEIGHT);
 static unsigned char *map;

 static unsigned int h_arr[93];
 static unsigned int s_arr[93];
 static unsigned int v_arr[93];

 static tRGBColor rgb;
 static tRGBColor rgb_old = {0, 0, 0};
 static tHSVColor hsv;

 map = (unsigned char *)cap.Capture();

 // clear statistic-arrays
 memset(&h_arr, 0, sizeof(h_arr));
 memset(&s_arr, 0, sizeof(s_arr));
 memset(&v_arr, 0, sizeof(v_arr));

 // scan picture-data
 for (int j = 0; j < HEIGHT; j++) 
 {
   for (int i = 0; i < WIDTH; i++) 
   {
     rgb.b = *map++;
     rgb.g = *map++;
     rgb.r = *map++;

     // filter black pixels
     if ((rgb.r <= 25) && (rgb.g <= 25) && (rgb.b <= 25)) { continue; }

     hsv = RGB2HSV(rgb);

     h_arr[(int)hsv.h]++;
     s_arr[(int)hsv.h] += hsv.s;
     v_arr[(int)hsv.h] = MAX(v_arr[(int)hsv.h], hsv.v);
   }
 }

 // find most used hue
 unsigned char max_ix = 0;
 for (int i=1; i<93; i++)
 {
   if (h_arr[max_ix] < h_arr[i]) { max_ix = i;}
 }

 hsv.h = max_ix;                          // most used hue
 if (0 != h_arr[max_ix])
 {
   hsv.s = s_arr[max_ix] / h_arr[max_ix]; // average saturation
   hsv.v = v_arr[max_ix]; // brightest value
 } else { // black screen
   hsv.s = 0;
   hsv.v = 0;
 }

 rgb = HSV2RGB(hsv);

 /* --- filtering --- */
 rgb.r = ((rgb.r * (AtmoSetup.PercentNew)) + (rgb_old.r * (100 - AtmoSetup.PercentNew))) / 100;
 rgb.g = ((rgb.g * (AtmoSetup.PercentNew)) + (rgb_old.g * (100 - AtmoSetup.PercentNew))) / 100;
 rgb.b = ((rgb.b * (AtmoSetup.PercentNew)) + (rgb_old.b * (100 - AtmoSetup.PercentNew))) / 100;
 rgb_old = rgb; // save filter values;   

 return rgb;
}

tHSVColor cAtmoThread::RGB2HSV(tRGBColor color)
{
 unsigned char min, max, delta;
 unsigned char s, v;
 float h;
 tHSVColor hsv;

 min = MIN(MIN(color.r, color.g), color.b);
 max = MAX(MAX(color.r, color.g), color.b);
 
 delta = max - min;

 v = max;

 if (v == 0) { s = 0; }
 else { s = delta * (255 / v); }

 if (s == 0) { h = 0.0f; } 
 else
 {
   if (max == color.r)
     { h = 60.0f * (color.g - color.b) / delta; }
   else if (max == color.g)
     { h = 120.0f + 60.0f * (color.b - color.r) / delta; }
   else // if (max == color.b)
     { h = 240.0f + 60.0f * (color.r - color.g) / delta; }

   if (h < 0.0f) { h += 360.0f; }

   // value range: 0..360 -> 0..92 (3 * 32 (0..31) colors)
   h = h * 92.0f / 360.0f;

   // value range: 0..255 -> 0..31
   s = (unsigned char)(s * 31.0f / 255.0f);

   // value range: 0..255 -> 0..31
   v = (unsigned char)(v * 31.0f / 255.0f);
 }

 hsv.h = (unsigned char)h;
 hsv.s = (unsigned char)s;
 hsv.v = (unsigned char)v;

 return hsv;
}

tRGBColor cAtmoThread::HSV2RGB(tHSVColor color)
{
 tRGBColor rgb = {0, 0, 0};

 if (color.v <= AtmoSetup.DarkLimit) return rgb;

 if (color.h <= 31) // between red and green
 {
   rgb.r = 31 - color.h;
   rgb.g = color.h;
   rgb.b = 0;
 }
 else if ((color.h >= 32) && (color.h <= 62)) // between green and blue
 {
   rgb.r = 0;
   rgb.g = 31 - (color.h - 31);
   rgb.b = color.h - 31;
 }
 else // if ((color.h >= 63) && (color.h <= 92)) // between blue and red
 {
   rgb.r = color.h - 62;
   rgb.g = 0;
   rgb.b = 31 - (color.h - 62);
 }

 // saturation offset;
 color.s += AtmoSetup.SatOffset; if (color.s > 31) color.s = 31;

 // saturation
 rgb.r += (31 - color.s); if (rgb.r > 31) rgb.r = 31;
 rgb.g += (31 - color.s); if (rgb.g > 31) rgb.g = 31;
 rgb.b += (31 - color.s); if (rgb.b > 31) rgb.b = 31;

 // white calibration
 if (rgb.r > (31 - AtmoSetup.CalibRed))   { rgb.r -= (31 - AtmoSetup.CalibRed); }   else { rgb.r = 0; }
 if (rgb.g > (31 - AtmoSetup.CalibGreen)) { rgb.g -= (31 - AtmoSetup.CalibGreen); } else { rgb.g = 0; }
 if (rgb.b > (31 - AtmoSetup.CalibBlue))  { rgb.b -= (31 - AtmoSetup.CalibBlue); }  else { rgb.b = 0; }
 
 return rgb;
}
